import unittest
from unittest.mock import AsyncMock
from ziplineio.dependency_injector import inject
from ziplineio.models import Request
from ziplineio.app import App, Router


class LoggingService:
    pass


LoggingService.log = AsyncMock()


class TestApp(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        # Initialize the app
        self.app = App()

        # Example middleware that just passes through the request
        async def example_middleware(req, **kwargs):
            return req, {"example_ctx": "some_value"}

        # Register middlewares for testing
        self.example_middleware = [example_middleware]

    async def test_basic_route(self):
        # Mock request data
        req = Request(method="GET", path="/")
        req.query_params = {"bar": "baz"}

        # Define a simple handler for testing
        @self.app.get("/")
        async def test_handler(req: Request, ctx: dict):
            return {"message": "Hello, world!"}

        # Call the route
        handler, params = self.app._router.get_handler("GET", "/")
        response = await handler(req, {})

        # Assertions
        self.assertEqual(response["message"], "Hello, world!")

    async def test_middleware_injection(self):
        # Mock request data
        req = Request(method="GET", path="/with-middleware")

        # Define a handler with middleware
        @self.app.get("/with-middleware")
        @self.app.middleware(self.example_middleware)
        async def test_handler_with_middleware(req: Request, ctx: dict):
            return {**ctx, "message": "Middleware test"}

        # Call the route
        handler, params = self.app._router.get_handler("GET", "/with-middleware")
        response = await handler(req)

        # Assertions
        self.assertEqual(response["message"], "Middleware test")
        self.assertEqual(response["example_ctx"], "some_value")

    async def test_dependency_injection(self):
        # Mock request data
        req = Request(method="GET", path="/with-middleware")

        # Define a handler with middleware
        @self.app.get("/")
        @inject(LoggingService, name="logger")
        async def test_handler_with_dependency(
            req: Request, ctx: dict, logger: LoggingService
        ):
            await logger.log("Dependency test")
            return {**ctx, "message": "Middleware test"}

        # Call the route
        handler, params = self.app._router.get_handler("GET", "/")
        await handler(req, ctx={})

        # Assertions

        LoggingService.log.assert_called_with("Dependency test")

    async def test_route_not_found(self):
        # Call the route
        handler, params = self.app._router.get_handler("GET", "/not-found")

        # Assertions
        self.assertIsNone(handler)

    async def test_post_route(self):
        # Mock request data for a POST route
        req = Request(method="POST", path="/submit")

        # Define a POST handler
        @self.app.post("/submit")
        async def post_handler(req: Request):
            return {"message": "Post received"}

        # Call the route
        handler, params = self.app._router.get_handler("POST", "/submit")
        response = await handler(req)

        # Assertions
        self.assertEqual(response["message"], "Post received")

    async def test_sub_route(self):
        # Mock request data for a POST route
        req = Request(method="GET", path="/submit")

        @self.app.get("/userform/submit")
        async def post_handler(req: Request):
            return {"message": "Post received"}

        # Call the route
        handler, params = self.app._router.get_handler("GET", "/userform/submit")
        response = await handler(req)

        # Assertions
        self.assertEqual(response["message"], "Post received")

    async def test_sub_routers(self):
        foo_router = Router("/b")

        @foo_router.get("/c")
        async def bar_handler(req: Request):
            return {"message": "Foo Bar received"}

        req = Request(method="GET", path="/a/b/c")

        self.app.router("/a", foo_router)

        # Call the route
        handler, params = self.app._router.get_handler("GET", "/a/b/c")
        response = await handler(req)

        # Assertions
        self.assertEqual(response["message"], "Foo Bar received")

    async def test_wildcard_route(self):
        # Mock request data for a POST route
        req = Request(method="GET", path="/user/123")

        @self.app.get("/user/:id")
        async def user_handler(req: Request):
            return {"message": f"User {req.path_params['id']} received"}

        # Call the route
        handler, params = self.app._router.get_handler("GET", "/user/123")
        req.path_params = params
        response = await handler(req)

        # Assertions
        self.assertEqual(response["message"], "User 123 received")


if __name__ == "__main__":
    unittest.main()



from typing import Generic, Protocol, TypeVar

from ziplineio.models import Request
from ziplineio.response import Response

T = TypeVar("T")


class Handler(Protocol, Generic[T]):
    async def __call__(self, req: Request, ctx: T) -> bytes | str | Response:
        pass



from typing import Any, Callable, List, Tuple, Type
import inspect
import asyncio
from ziplineio.dependency_injector import inject, injector, DependencyInjector
from ziplineio.exception import BaseHttpException
from ziplineio.handler import Handler
from ziplineio.models import Request
from ziplineio.response import RawResponse, format_response
from ziplineio.router import Router
from ziplineio.utils import parse_scope


class App:
    _router: Router
    _injector: DependencyInjector

    def __init__(self) -> None:
        self._router = Router()
        self._injector = injector

    def router(self, prefix: str, router: Router) -> None:
        self._router.add_sub_router(prefix, router)

    def get(self, path: str) -> Callable[[Handler], Callable]:
        return self._router.get(path)

    def post(self, path: str) -> Callable[[Handler], Callable]:
        return self._router.post(path)

    def put(self, path: str) -> Callable[[Handler], Callable]:
        return self._router.put(path)

    def delete(self, path: str) -> Callable[[Handler], Callable]:
        return self._router.delete(path)

    def get_handler(self, method: str, path: str) -> Tuple[Handler, dict]:
        app_level_deps = self._injector._injected_services["app"]
        handler, params = self._router.get_handler(method, path)

        # TODO: Inject deps when added to the router, not here
        # inject app-level dependencies into the handler
        for name, service in app_level_deps.items():
            print(f"Service: {service}")
            handler = inject(service, name)(handler)

        return handler, params

    def inject(
        self, service_class: Type, name: str = None
    ) -> Callable[[Callable], Callable]:
        return self._injector.add_injected_service(service_class, name, "app")

    def middleware(self, middlewares: List[Callable]) -> Callable[[Callable], Callable]:
        def decorator(handler: Callable) -> Callable:
            async def wrapped_handler(req: Request, **kwargs):
                # Ensure 'ctx' is always present in kwargs
                print(f"Kwargs: {kwargs}")
                kwargs.setdefault("ctx", {})

                for middleware in middlewares:
                    try:
                        # if the middleware func takes params, pass them in. Otherwise, just pass req
                        if len(middleware.__code__.co_varnames) > 1:
                            _res = await middleware(req, **kwargs)
                        else:
                            _res = await middleware(req)

                        if len(_res) != 2:
                            req = _res
                        else:
                            req, middleware_ctx = _res

                        kwargs["ctx"].update(middleware_ctx)
                    except Exception as e:
                        return {
                            "status": 500,
                            "headers": [],
                            "body": b"Internal server error: " + str(e).encode(),
                        }

                    # check if the middleware returned a response
                    if not isinstance(req, Request):
                        return req

                # Call the final handler with the updated kwargs including ctx
                return await handler(req, **kwargs)

            return wrapped_handler

        return decorator

    async def call_handler(self, handler: Handler, req: Request) -> RawResponse:
        try:
            if not inspect.iscoroutinefunction(handler):
                response = await asyncio.to_thread(handler, req)
            else:
                response = await handler(req)
            print(f"Response in call_handler: {response}")
            print(f"Response type: {type(response)}")

        except BaseHttpException as e:
            response = e
        except Exception as e:
            response = BaseHttpException(e, 500)

        print(f"Response: {response}")

        return format_response(response)

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        async def uvicorn_handler(scope: dict, receive: Any, send: Any) -> None:
            if scope["type"] == "http":
                req = parse_scope(scope)
                handler, path_params = self.get_handler(req.method, req.path)
                if handler is None:
                    response = {"status": 404, "headers": [], "body": b"Not found"}
                else:
                    req.path_params = path_params
                    response = await self.call_handler(handler, req)

                await send(
                    {
                        "type": "http.response.start",
                        "status": response["status"],
                        "headers": response["headers"],
                    }
                )

                await send(
                    {
                        "type": "http.response.body",
                        "body": response["body"],
                    }
                )

        return uvicorn_handler


app = App()


async def logging_middleware(req: Request) -> Tuple[Request, dict]:
    print(f"Received request: {req.method} {req.path}")
    # Pass along request and an empty context (no modifications)
    return req, {}


async def auth_middleware(req: Request) -> Tuple[Request, dict]:
    # Modify the context with some auth information
    return req, {"auth": "Unauthorized"}


class LoggingService:
    def log(self, message: str) -> None:
        print(f"Logging: {message}")


@app.get("/")
@inject(LoggingService)
@app.middleware([logging_middleware, auth_middleware])
async def handler(req: Request, ctx: dict, LoggingService: LoggingService) -> dict:
    res = {"message": "Hello, world!"}

    # Use the injected LoggingService
    LoggingService.log(f"Params: {req.query_params.get('bar')}")

    return {**res, **ctx}


@app.get("/foo/:bar")
async def handler2(req: Request) -> dict:
    bar = req.path_params.get("bar")
    res = {"message": f"Hello, {bar}!"}

    return res



import unittest

from ziplineio.app import App
from ziplineio.router import Router
from ziplineio.dependency_injector import DependencyInjector, inject
import ziplineio.dependency_injector


class MockService:
    pass


class TestDependencyInjection(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        # Initialize the ziplineio
        self.ziplineio = App()

        # reset the injector
        ziplineio.dependency_injector.injector = DependencyInjector()
        self.ziplineio._injector = ziplineio.dependency_injector.injector

    async def test_inject_service(self):
        # Define a service to inject
        class Service:
            def __init__(self):
                self.value = "Service value"

        # Inject the service
        @inject(Service)
        async def test_handler(req, service: Service):
            return {"message": service.value}

        # Call the handler
        response = await test_handler({})
        self.assertEqual(response["message"], "Service value")

    async def test_multiple_inject_same_instance(self):
        # Define a service to inject
        class Service:
            def __init__(self):
                self.value = "Service value"

        # Inject the service
        @inject(Service, name="service")
        async def test_handler1(req, service: Service):
            return {"message": service.value}

        # Inject the service again
        @inject(Service, name="service")
        async def test_handler2(req, service: Service):
            return {"message": service.value}

        # expect 1 instance of the service
        self.assertEqual(len(self.ziplineio._injector._injected_services["func"]), 1)

    async def inject_without_params(self):
        # Define a service to inject
        class Service:
            def __init__(self):
                self.value = "Service value"

        # Inject the service
        @inject(Service)
        async def test_handler(req):
            return {"message": "No service"}

        # Call the handler; should not raise an error
        response = await test_handler({})
        self.assertEqual(response["message"], "No service")

    async def test_inject_service_with_name(self):
        # Define a service to inject
        class Service:
            def __init__(self):
                self.value = "Service value"

        # Inject the service
        @inject(Service, name="my_service")
        async def test_handler(req, my_service: Service):
            return {"message": my_service.value}

        # Call the handler
        response = await test_handler({})
        self.assertEqual(response["message"], "Service value")

    async def test_inject_app_level_service(self):
        # Inject the service at the ziplineio level

        self.ziplineio.inject(MockService, name="service")

        @self.ziplineio.get("/")
        async def test_handler1(req, service: MockService):
            return {"message": f"Service {MockService} injected"}

        # doesn't throw an error
        @self.ziplineio.get("/no-service")
        async def test_handler2(req):
            return {"message": "No service injected"}

        # Call the handler
        handler, params = self.ziplineio.get_handler("GET", "/")
        response = await handler({})
        self.assertEqual(response["message"], f"Service {MockService} injected")

    async def test_inject_router_level(self):
        # Define a service to inject
        class Service:
            def __init__(self):
                self.value = "Service"

        # Create a router
        router = Router()

        # Inject the service at the router level
        router.inject(Service, name="service")

        @router.get("/")
        async def test_handler(req, service: Service):
            return {"message": service.value}

        # Call the handler
        handler, params = router.get_handler("GET", "/")
        response = await handler({})
        self.assertEqual(response["message"], "Service")
